//-------------------------------------------------------------------------------------------------+
// LocalMap.cpp
// Ǐn}RTC̃CNX̊֐`
//-------------------------------------------------------------------------------------------------+

#include "LocalMap.h"
#include <iostream>
#include <mmsystem.h>

using namespace std;


//-------------------------------------------------------------------------------------------------+
// Module specification
//-------------------------------------------------------------------------------------------------+
static const char* localmap_spec[] =
{
    "implementation_id", "LocalMap",
    "type_name",         "LocalMap",
    "description",       "Generate Local Map",
    "version",           "1.0.0",
    "vendor",            "aisl.cs.tut.ac.jp",
    "category",          "mapping",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.LocalMapWidth", "200",
    "conf.default.LocalMapHeight", "200",
    "conf.default.LocalMapScale", "0.05",
    "conf.default.SensorPositionX_m", "0.0",
    "conf.default.SensorPositionY_m", "0.0",
    "conf.default.SensorMaxDistance_mm", "30000.0",
    "conf.default.SensorMinDistance_mm", "20.0",
    "conf.default.Error_code", "19",
    "conf.default.Error_mode", "0",
    "conf.default.SensorPositionTheta_deg", "0.0",
    // Widget
    "conf.__widget__.LocalMapWidth", "text",
    "conf.__widget__.LocalMapHeight", "text",
    "conf.__widget__.LocalMapScale", "text",
    "conf.__widget__.SensorPositionX_m", "text",
    "conf.__widget__.SensorPositionY_m", "text",
    "conf.__widget__.SensorMaxDistance_mm", "text",
    "conf.__widget__.SensorMinDistance_mm", "text",
    "conf.__widget__.Error_code", "text",
    "conf.__widget__.Error_mode", "text",
    "conf.__widget__.SensorPositionTheta_deg", "text",
    // Constraints
    ""
};


//-------------------------------------------------------------------------------------------------+
// RXgN^
//-------------------------------------------------------------------------------------------------+
LocalMap::LocalMap(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_rangedataIn("rangedata", m_rangedata),
    m_poseIn("pose", m_pose),
    m_RelativeMapServicePort("RelativeMapService"),
	m_RelativeMapService( &localmap, &g_critSec )
{

	cerr << "------------------------------------------------------------+\n"
		 << " LocalMap Component (Ver. 13th July 2011)\n"
		 << "------------------------------------------------------------+\n\n";

	InitializeCriticalSection( &g_critSec );
}


//-------------------------------------------------------------------------------------------------+
// fXgN^
//-------------------------------------------------------------------------------------------------+
LocalMap::~LocalMap(){

	delete LocalMap_A;
}


//-------------------------------------------------------------------------------------------------+
// vON̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t LocalMap::onInitialize(){

	// Set InPort buffers
	addInPort("rangedata", m_rangedataIn);
	addInPort("pose", m_poseIn);

	// Set service provider to Ports
	m_RelativeMapServicePort.registerProvider("RelativeMapService", "MRFC::RelativeMapService", m_RelativeMapService);

	// Set CORBA Service Ports
	addPort(m_RelativeMapServicePort);

	// Bind variables and configuration variable
	bindParameter("LocalMapWidth", m_LocalMapWidth, "200");
	bindParameter("LocalMapHeight", m_LocalMapHeight, "200");
	bindParameter("LocalMapScale", m_LocalMapScale, "0.05");
	bindParameter("SensorPositionX_m", m_SensorX, "0.0");
	bindParameter("SensorPositionY_m", m_SensorY, "0.0");
	bindParameter("SensorMaxDistance_mm", m_SensorMaxDistance, "30000.0");
	bindParameter("SensorMinDistance_mm", m_SensorMinDistance, "20.0");
	bindParameter("Error_code", m_error_code, "19");
	bindParameter("Error_mode", m_Error_mode, "0");
	bindParameter("SensorPositionTheta_deg", m_SensorTheta, "0.0");

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// ̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t LocalMap::onActivated(RTC::UniqueId ec_id){

	int x_distance = (int)( 0.5 * m_LocalMapWidth  * m_LocalMapScale * 1000.0 );		// {bgOmm܂Œn}ɓ邩
	int y_distance = (int)( 0.5 * m_LocalMapHeight * m_LocalMapScale * 1000.0 );		// {bg牡mm܂Œn}ɓ邩

	LocalMap_A = new LPM( -x_distance, x_distance, -y_distance, y_distance, (int)( m_LocalMapScale * 1000.0 ));

	set_old_pose = false;

	LocalMap_A->set_pose( 0.0, 0.0, 0.0 );

	EnterCriticalSection( &g_critSec );

	InitLocalMap();

	localmap.pose.position.x = 0.0;
	localmap.pose.position.y = 0.0;
	localmap.pose.heading    = 0.0;

	localmap.tm.sec  = 0;
	localmap.tm.nsec = 0;

	for( int i=0; i<(int)localmap.cells.length(); i++ ) localmap.cells[i] = -1;

	LeaveCriticalSection( &g_critSec );

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// 񊈐̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t LocalMap::onDeactivated(RTC::UniqueId ec_id){

	EnterCriticalSection( &g_critSec );

	delete LocalMap_A;
	LocalMap_A = 0;

	LeaveCriticalSection( &g_critSec );

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// sRTC̃C[`
// Ǐn}̍XVs
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t LocalMap::onExecute(RTC::UniqueId ec_id){

	if( !m_poseIn.isNew() || !m_rangedataIn.isNew() ) return RTC::RTC_OK;

	while( !m_rangedataIn.isEmpty() ) m_rangedataIn.read();
	while( !m_poseIn.isEmpty() ) m_poseIn.read();

	UpdateLocalMap();
	MakeLocalMap();

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// Ǐn}̍XV
//-------------------------------------------------------------------------------------------------+
void LocalMap::UpdateLocalMap(){

	Pose Current_Pose;

	Current_Pose.x     = m_pose.data.position.x * 1000;		// Pʌnϊ [m]->[mm]
	Current_Pose.y     = m_pose.data.position.y * 1000;		// Pʌnϊ [m]->[mm]
	Current_Pose.theta = m_pose.data.heading;

	LocalMap_A->set_pose( Current_Pose.x, Current_Pose.y, Current_Pose.theta );


	Pose Pose_diff;		// p̍vZp

	if( !set_old_pose ){

		Pose_diff.x     = 0.0;
		Pose_diff.y     = 0.0;
		Pose_diff.theta = 0.0;
	}
	else{
		double dx     = Current_Pose.x     - Old_Pose.x;
		double dy     = Current_Pose.y     - Old_Pose.y;
		double dtheta = Current_Pose.theta - Old_Pose.theta;

		Pose_diff.x     =  cos(Old_Pose.theta) * dx + sin(Old_Pose.theta) * dy;
		Pose_diff.y     = -sin(Old_Pose.theta) * dx + cos(Old_Pose.theta) * dy;
		Pose_diff.theta = dtheta;
	}

	Old_Pose.x     = Current_Pose.x;
	Old_Pose.y     = Current_Pose.y;
	Old_Pose.theta = Current_Pose.theta;
	set_old_pose   = true;


	Pose Sensor_Pose;

	// Pʌnϊ [m]->[mm] [deg]->[rad]
	Sensor_Pose.x     = m_SensorX * 1000;
	Sensor_Pose.y     = m_SensorY * 1000;
	Sensor_Pose.theta = m_SensorTheta / 180.0 * M_PI;

	double *dist = StoreRangeData();

	// n}̍XV
	LocalMap_A->MapUpdate(
		Pose_diff, dist,
		m_rangedata.data.startPosition-90,
		m_rangedata.data.endPosition-90,
		m_rangedata.data.distance.length(),
		m_SensorMinDistance,
		m_SensorMaxDistance,
		Sensor_Pose
	);

	delete [] dist;
}


//------------------------------------------------------------------------------------------------+
// Ǐn}̐
//------------------------------------------------------------------------------------------------+
void LocalMap::MakeLocalMap(){

	EnterCriticalSection( &g_critSec );

	InitLocalMap();

	localmap.pose.position.x = m_pose.data.position.x;
	localmap.pose.position.y = m_pose.data.position.y;
	localmap.pose.heading    = m_pose.data.heading;

	localmap.tm = m_pose.tm;

	int index = 0;

	for( int y=0; y<localmap.mapconfig.height; y++ ) for( int x=0; x<localmap.mapconfig.width; x++ ){

		localmap.cells[index] = (int)( LocalMap_A->map[x][y] * 100.0 );
		index++;
	}

	LeaveCriticalSection( &g_critSec );
}


//------------------------------------------------------------------------------------------------+
// Ǐn}𐶐čŏ̏s
// ̊֐ĂԑOŃNeBJZNV̏Ăяoׂ
//------------------------------------------------------------------------------------------------+
void LocalMap::InitLocalMap(){

	localmap = TimedRelativeOGMapData();

	localmap.mapconfig.width  = m_LocalMapWidth;
	localmap.mapconfig.height = m_LocalMapHeight;
	localmap.mapconfig.xScale = m_LocalMapScale;
	localmap.mapconfig.yScale = m_LocalMapScale;

	localmap.cells.length( m_LocalMapWidth * m_LocalMapHeight );

	localmap.mapconfig.origin.heading    = 0.0;
	localmap.mapconfig.origin.position.x = -m_LocalMapScale * m_LocalMapWidth  / 2.0;
	localmap.mapconfig.origin.position.y = -m_LocalMapScale * m_LocalMapHeight / 2.0;
}


//-------------------------------------------------------------------------------------------------+
// ZTf[^zɎ荞
// z͌ĂяoŊJ
//-------------------------------------------------------------------------------------------------+
double *LocalMap::StoreRangeData(){

	double *dist = new double[m_rangedata.data.distance.length()];

	for( int i=0; i<m_rangedata.data.distance.length(); i++ ){

		if( m_Error_mode == 1 ){

			if( m_rangedata.data.distance[i] <= 20 ) dist[i] = m_SensorMaxDistance + 100;
			else if( m_rangedata.data.distance[i] > 20 ) dist[i] = m_rangedata.data.distance[i];
			else dist[i] = -1;
		}
		else if( m_Error_mode == 2 ){

			if( m_rangedata.data.distance[i] <= 5 ) dist[i] = m_SensorMaxDistance + 100;
			else if( m_rangedata.data.distance[i] > 5 ) dist[i] = m_rangedata.data.distance[i];
			else dist[i] = -1;
		}
		else{
			if( m_rangedata.data.distance[i] == m_error_code ) dist[i] = m_SensorMaxDistance + 100;
			else if( m_rangedata.data.distance[i] > m_SensorMinDistance ) dist[i] = m_rangedata.data.distance[i];
			else dist[i] = -1;
		}
	}

	return dist;
}



//-------------------------------------------------------------------------------------------------+
extern "C"
{
 
  void LocalMapInit(RTC::Manager* manager)
  {
    coil::Properties profile(localmap_spec);
    manager->registerFactory(profile,
                             RTC::Create<LocalMap>,
                             RTC::Delete<LocalMap>);
  }

};
